import asyncio
import gc
import io
import mimetypes
import urllib.parse
from pathlib import PurePosixPath
from time import perf_counter
from urllib.parse import unquote
from wsgiref.handlers import format_date_time

import av
import fitz  # PyMuPDF
import numpy as np
import pillow_heif
from PIL import Image
from sanic import Blueprint, empty, raw, redirect
from sanic.exceptions import NotFound
from sanic.log import logger

from cista import auth, config
from cista.util.filename import sanitize

pillow_heif.register_heif_opener()

bp = Blueprint("preview", url_prefix="/preview")


@bp.on_request
async def verify_preview(request):
    """Verify access to preview routes."""
    await auth.verify(request)


# Map EXIF Orientation value to a corresponding PIL transpose
EXIF_ORI = {
    2: Image.Transpose.FLIP_LEFT_RIGHT,
    3: Image.Transpose.ROTATE_180,
    4: Image.Transpose.FLIP_TOP_BOTTOM,
    5: Image.Transpose.TRANSPOSE,
    6: Image.Transpose.ROTATE_270,
    7: Image.Transpose.TRANSVERSE,
    8: Image.Transpose.ROTATE_90,
}


@bp.get("/<path:path>")
async def preview(req, path):
    """Preview a file"""
    maxsize = int(req.args.get("px", 1024))
    maxzoom = float(req.args.get("zoom", 2.0))
    quality = int(req.args.get("q", 60))
    rel = PurePosixPath(sanitize(unquote(path)))
    filepath = config.config.path / rel
    stat = filepath.lstat()
    etag = config.derived_secret(
        "preview", rel, stat.st_mtime_ns, quality, maxsize, maxzoom
    ).hex()
    savename = PurePosixPath(filepath.name).with_suffix(".avif")
    headers = {
        "etag": etag,
        "last-modified": format_date_time(stat.st_mtime),
        "cache-control": "max-age=604800, immutable"
        + ("" if config.config.public else ", private"),
        "content-type": "image/avif",
        "content-disposition": f"inline; filename*=UTF-8''{urllib.parse.quote(savename.as_posix())}",
    }
    if req.headers.if_none_match == etag:
        # The client has it cached, respond 304 Not Modified
        return empty(304, headers=headers)

    if not filepath.is_file():
        raise NotFound("File not found")

    img = await asyncio.get_event_loop().run_in_executor(
        req.app.ctx.threadexec, dispatch, filepath, quality, maxsize, maxzoom
    )
    if not img:
        # Preview generation failed, redirect to the file itself
        return redirect(f"/files/{path}", status=303)
    return raw(img, headers=headers)


def dispatch(path, quality, maxsize, maxzoom):
    try:
        if path.suffix.lower() in (".pdf", ".xps", ".epub", ".mobi"):
            return process_pdf(path, quality=quality, maxsize=maxsize, maxzoom=maxzoom)
        type, _ = mimetypes.guess_type(path.name)
        if type and type.startswith("video/"):
            return process_video(path, quality=quality, maxsize=maxsize)
        return process_image(path, quality=quality, maxsize=maxsize)
    except ValueError as e:
        logger.warning(f"Cannot generate preview for {path}: {e}")
    except Exception as e:
        logger.exception(f"Error generating preview for {path}: {e}")


def process_image(path, *, maxsize, quality):
    t_load = perf_counter()
    with Image.open(path) as img:
        # Force decode to include I/O in load timing
        img.load()
        t_proc = perf_counter()
        # Resize
        w, h = img.size
        img.thumbnail((min(w, maxsize), min(h, maxsize)))
        # Transpose pixels according to EXIF Orientation
        orientation = img.getexif().get(274, 1)
        if orientation in EXIF_ORI:
            img = img.transpose(EXIF_ORI[orientation])
        # Save as AVIF
        imgdata = io.BytesIO()
        t_save = perf_counter()
        img.save(imgdata, format="avif", quality=quality, speed=10, max_threads=1)

    t_end = perf_counter()
    ret = imgdata.getvalue()

    load_ms = (t_proc - t_load) * 1000
    proc_ms = (t_save - t_proc) * 1000
    save_ms = (t_end - t_save) * 1000
    logger.debug(
        "Preview image %s: load=%.1fms process=%.1fms save=%.1fms",
        path.name,
        load_ms,
        proc_ms,
        save_ms,
    )

    return ret


def process_pdf(path, *, maxsize, maxzoom, quality, page_number=0):
    t_load_start = perf_counter()
    pdf = fitz.open(path)
    page = pdf.load_page(page_number)
    w, h = page.rect[2:4]
    zoom = min(maxsize / w, maxsize / h, maxzoom)
    mat = fitz.Matrix(zoom, zoom)
    pix = page.get_pixmap(matrix=mat)
    t_load_end = perf_counter()

    t_save_start = perf_counter()
    ret = pix.pil_tobytes(format="avif", quality=quality, speed=10, max_threads=1)
    t_save_end = perf_counter()

    logger.debug(
        "Preview pdf %s: load+render=%.1fms save=%.1fms",
        path.name,
        (t_load_end - t_load_start) * 1000,
        (t_save_end - t_save_start) * 1000,
    )
    return ret


def process_video(path, *, maxsize, quality):
    frame = None
    imgdata = io.BytesIO()
    istream = ostream = icc = occ = frame = None
    t_load_start = perf_counter()
    # Initialize to avoid "possibly unbound" in static analysis when exceptions occur
    t_load_end = t_load_start
    t_save_start = t_load_start
    t_save_end = t_load_start
    with (
        av.open(str(path)) as icontainer,
        av.open(imgdata, "w", format="avif") as ocontainer,
    ):
        istream = icontainer.streams.video[0]
        istream.codec_context.skip_frame = "NONKEY"
        icontainer.seek((icontainer.duration or 0) // 8)
        for frame in icontainer.decode(istream):
            if frame.dts is not None:
                break
        else:
            raise RuntimeError("No frames found in video")

        # Resize frame to thumbnail size
        if frame.width > maxsize or frame.height > maxsize:
            scale_factor = min(maxsize / frame.width, maxsize / frame.height)
            new_width = int(frame.width * scale_factor)
            new_height = int(frame.height * scale_factor)
            frame = frame.reformat(width=new_width, height=new_height)

        # Apply EXIF rotation if present
        if frame.rotation:
            # frame.rotation indicates clockwise rotation needed to display correctly
            # np.rot90 rotates counter-clockwise, so we negate k
            k = (frame.rotation // 90) % 4  # Convert to counter-clockwise rotations
            if k == 2:
                # 180° rotation can be done in YUV420p, preserving HDR
                try:
                    fplanes = frame.to_ndarray()
                    # Split into Y, U, V planes of proper dimensions
                    planes = [
                        fplanes[: frame.height],
                        fplanes[
                            frame.height : frame.height + frame.height // 4
                        ].reshape(frame.height // 2, frame.width // 2),
                        fplanes[frame.height + frame.height // 4 :].reshape(
                            frame.height // 2, frame.width // 2
                        ),
                    ]
                    # Rotate each plane by 180°
                    planes = [np.rot90(p, 2) for p in planes]
                    # Restore PyAV format
                    planes = np.hstack([p.flat for p in planes]).reshape(
                        -1, planes[0].shape[1]
                    )
                    frame = av.VideoFrame.from_ndarray(planes, format=frame.format.name)
                    del planes, fplanes
                except Exception as e:
                    logger.exception(f"Error rotating video frame by 180°: {e}")
            elif k in (1, 3):
                # 90° or 270° rotation requires RGB conversion (loses HDR)
                try:
                    rgb = frame.to_ndarray(format="rgb24")
                    rgb = np.rot90(rgb, k)
                    frame = av.VideoFrame.from_ndarray(rgb, format="rgb24")
                    frame = frame.reformat(
                        format="yuv420p"
                    )  # Convert back for encoding
                    del rgb
                except Exception as e:
                    logger.exception(
                        f"Error rotating video frame by {frame.rotation}°: {e}"
                    )
        t_load_end = perf_counter()

        t_save_start = perf_counter()
        crf = str(int(63 * (1 - quality / 100) ** 2))  # Closely matching PIL quality-%
        ostream = ocontainer.add_stream(
            "av1",
            options={
                "crf": crf,
                "usage": "realtime",
                "cpu-used": "8",
                "threads": "1",
            },
        )
        assert isinstance(ostream, av.VideoStream)
        ostream.width = frame.width
        ostream.height = frame.height
        ostream.pix_fmt = frame.format.name
        icc = istream.codec_context
        occ = ostream.codec_context

        # Copy HDR metadata from input video stream
        occ.color_primaries = icc.color_primaries
        occ.color_trc = icc.color_trc
        occ.colorspace = icc.colorspace
        occ.color_range = icc.color_range

        ocontainer.mux(ostream.encode(frame))
        ocontainer.mux(ostream.encode(None))  # Flush the stream
        t_save_end = perf_counter()

    # Capture frame dimensions before cleanup
    ret = imgdata.getvalue()
    logger.debug(
        "Preview video %s: load+decode=%.1fms save=%.1fms",
        path.name,
        (t_load_end - t_load_start) * 1000,
        (t_save_end - t_save_start) * 1000,
    )
    del imgdata, istream, ostream, icc, occ, frame
    gc.collect()
    return ret



import asyncio
import queue
import shutil
import sys
import threading
import time
from contextlib import suppress
from os import stat_result
from pathlib import Path, PurePosixPath
from stat import S_ISDIR, S_ISREG

import msgspec
from natsort import humansorted, natsort_keygen, ns
from sanic.log import logger

from cista import config
from cista.fileio import fuid
from cista.protocol import FileEntry, Space, UpdDel, UpdIns, UpdKeep

pubsub = {}
sortkey = natsort_keygen(alg=ns.LOCALE)


class State:
    def __init__(self):
        self.lock = threading.RLock()
        self._space = Space(0, 0, 0, 0, 0)
        self.root: list[FileEntry] = []

    @property
    def space(self):
        with self.lock:
            return self._space

    @space.setter
    def space(self, space):
        with self.lock:
            self._space = space


def treeiter(rootmod):
    relpath = PurePosixPath()
    for i, entry in enumerate(rootmod):
        if entry.level > 0:
            relpath = PurePosixPath(*relpath.parts[: entry.level - 1]) / entry.name
        yield i, relpath, entry


def treeget(rootmod: list[FileEntry], path: PurePosixPath):
    begin = None
    ret = []

    for i, relpath, entry in treeiter(rootmod):
        if begin is None:
            if relpath == path:
                begin = i
                ret.append(entry)
            continue
        if entry.level <= len(path.parts):
            break
        ret.append(entry)

    return begin, ret


def treeinspos(rootmod: list[FileEntry], relpath: PurePosixPath, relfile: int):
    # Find the first entry greater than the new one
    # precondition: the new entry doesn't exist
    isfile = 0
    level = 0
    i = 0
    for i, rel, entry in treeiter(rootmod):
        if entry.level > level:
            # We haven't found item at level, skip subdirectories
            continue
        if entry.level < level:
            # We have passed the level, so the new item is the first
            return i
        if level == 0:
            # root
            level += 1
            continue

        ename = rel.parts[level - 1]
        name = relpath.parts[level - 1]

        esort = sortkey(ename)
        nsort = sortkey(name)
        # Non-leaf are always folders, only use relfile at leaf
        isfile = relfile if len(relpath.parts) == level else 0

        # First compare by isfile, then by sorting order and if that too matches then case sensitive
        cmp = (
            entry.isfile - isfile
            or (esort > nsort) - (esort < nsort)
            or (ename > name) - (ename < name)
        )

        if cmp > 0:
            return i
        if cmp < 0:
            continue

        level += 1
        if level > len(relpath.parts):
            logger.error(
                f"insertpos level overflow: relpath={relpath}, i={i}, entry.name={entry.name}, entry.level={entry.level}, level={level}"
            )
            break
    else:
        i += 1

    return i


state = State()
rootpath: Path = None  # type: ignore
quit = threading.Event()

# Thread-safe queue for signaling path updates from websockets
_update_queue: queue.Queue[PurePosixPath] = queue.Queue()


def notify_change(*paths: PurePosixPath | str):
    """Signal that paths have changed. Called from control/upload websockets."""
    for path in paths:
        if isinstance(path, str):
            path = PurePosixPath(path)
        # Convert absolute paths to relative (strip leading /)
        if path.is_absolute():
            path = (
                PurePosixPath(*path.parts[1:])
                if len(path.parts) > 1
                else PurePosixPath()
            )
        # Skip root paths (empty, '.') to avoid full tree walks
        if not path.parts or path.parts == (".",):
            continue
        _update_queue.put(path)


## Filesystem scanning


def walk(rel: PurePosixPath, stat: stat_result | None = None) -> list[FileEntry]:
    path = rootpath / rel
    ret = []
    try:
        st = stat or path.stat()
        isfile = int(not S_ISDIR(st.st_mode))
        # st_blocks is in 512-byte units
        allocated = st.st_blocks * 512 if isfile else 0
        entry = FileEntry(
            level=len(rel.parts),
            name=rel.name,
            key=fuid(st),
            mtime=int(st.st_mtime),
            size=st.st_size if isfile else 0,
            allocated=allocated,
            isfile=isfile,
        )
        if isfile:
            return [entry]
        # Walk all entries of the directory
        ret: list[FileEntry] = [...]  # type: ignore
        li = []
        for f in path.iterdir():
            if quit.is_set():
                raise SystemExit("quit")
            if f.name.startswith("."):
                continue  # No dotfiles
            with suppress(FileNotFoundError):
                s = f.lstat()
                isfile = S_ISREG(s.st_mode)
                isdir = S_ISDIR(s.st_mode)
                if not isfile and not isdir:
                    continue
                li.append((int(isfile), f.name, s))
        # Build the tree as a list of FileEntries
        for [_, name, s] in humansorted(li):
            sub = walk(rel / name, stat=s)
            child = sub[0]
            entry = FileEntry(
                level=entry.level,
                name=entry.name,
                key=entry.key,
                mtime=max(entry.mtime, child.mtime),
                size=entry.size + child.size,
                allocated=entry.allocated + child.allocated,
                isfile=entry.isfile,
            )
            ret.extend(sub)
    except FileNotFoundError:
        pass  # Things may be rapidly in motion
    except OSError as e:
        if e.errno == 13:  # Permission denied
            pass
        logger.error(f"Watching {path=}: {e!r}")
    if ret:
        ret[0] = entry
    return ret


def update_root(loop):
    """Full filesystem scan"""
    old = state.root
    new = walk(PurePosixPath())
    if old != new:
        update = format_update(old, new)
        with state.lock:
            broadcast(update, loop)
            state.root = new


def update_path(rootmod: list[FileEntry], relpath: PurePosixPath, loop):
    """Called on FS updates, check the filesystem and broadcast any changes."""
    new = walk(relpath)
    obegin, old = treeget(rootmod, relpath)

    if old == new:
        return

    if obegin is not None:
        del rootmod[obegin : obegin + len(old)]

    if new:
        i = treeinspos(rootmod, relpath, new[0].isfile)
        rootmod[i:i] = new


def update_space(loop):
    """Called periodically to update the disk usage."""
    du = shutil.disk_usage(rootpath)
    root = state.root[0]
    space = Space(
        disk=du.total,
        free=du.free,
        used=du.used,
        storage=root.size,
        allocated=root.allocated,
    )
    # Update only on difference above 1 MB
    tol = 10**6
    old = msgspec.structs.astuple(state.space)
    new = msgspec.structs.astuple(space)
    if any(abs(o - n) > tol for o, n in zip(old, new, strict=True)):
        state.space = space
        broadcast(format_space(space), loop)


## Messaging


def format_update(old, new):
    # Make keep/del/insert diff until one of the lists ends
    oidx, nidx = 0, 0
    oremain, nremain = set(old), set(new)
    update = []
    keep_count = 0
    iteration_count = 0
    # Precompute index maps to allow deterministic tie-breaking when both
    # candidates exist in both sequences but are not equal (rename/move cases)
    old_pos = {e: i for i, e in enumerate(old)}
    new_pos = {e: i for i, e in enumerate(new)}

    while oidx < len(old) and nidx < len(new):
        iteration_count += 1

        # Emergency brake for potential infinite loops
        if iteration_count > 50000:
            logger.error(
                f"format_update potential infinite loop! iteration={iteration_count}, oidx={oidx}, nidx={nidx}"
            )
            raise Exception(
                f"format_update infinite loop detected at iteration {iteration_count}"
            )

        modified = False
        # Matching entries are kept
        if old[oidx] == new[nidx]:
            entry = old[oidx]
            oremain.discard(entry)
            nremain.discard(entry)
            keep_count += 1
            oidx += 1
            nidx += 1
            continue

        if keep_count > 0:
            modified = True
            update.append(UpdKeep(keep_count))
            keep_count = 0

        # Items only in old are deleted
        del_count = 0
        while oidx < len(old) and old[oidx] not in nremain:
            oremain.remove(old[oidx])
            del_count += 1
            oidx += 1
        if del_count:
            update.append(UpdDel(del_count))
            continue

        # Items only in new are inserted
        insert_items = []
        while nidx < len(new) and new[nidx] not in oremain:
            entry = new[nidx]
            nremain.discard(entry)
            insert_items.append(entry)
            nidx += 1
        if insert_items:
            modified = True
            update.append(UpdIns(insert_items))

        if not modified:
            # Tie-break: both items exist in both lists but don't match here.
            # Decide whether to delete old[oidx] first or insert new[nidx] first
            # based on which alignment is closer.
            if oidx >= len(old) or nidx >= len(new):
                break
            cur_old = old[oidx]
            cur_new = new[nidx]

            pos_old_in_new = new_pos.get(cur_old)
            pos_new_in_old = old_pos.get(cur_new)

            # Default distances if not present (shouldn't happen if in remain sets)
            dist_del = (pos_old_in_new - nidx) if pos_old_in_new is not None else 1
            dist_ins = (pos_new_in_old - oidx) if pos_new_in_old is not None else 1

            # Prefer the operation with smaller forward distance; tie => delete
            if dist_del <= dist_ins:
                # Delete current old item
                oremain.discard(cur_old)
                update.append(UpdDel(1))
                oidx += 1
            else:
                # Insert current new item
                nremain.discard(cur_new)
                update.append(UpdIns([cur_new]))
                nidx += 1

    # Diff any remaining
    if keep_count > 0:
        update.append(UpdKeep(keep_count))
    if oremain:
        update.append(UpdDel(len(oremain)))
    elif nremain:
        update.append(UpdIns(new[nidx:]))

    return msgspec.json.encode({"update": update}).decode()


def format_space(usage):
    return msgspec.json.encode({"space": usage}).decode()


def format_root(root):
    return msgspec.json.encode({"root": root}).decode()


def broadcast(msg, loop):
    fut = asyncio.run_coroutine_threadsafe(abroadcast(msg), loop)
    return fut.result()


async def abroadcast(msg):
    client_count = 0
    try:
        for queue in pubsub.values():
            queue.put_nowait(msg)
            client_count += 1
    except Exception:
        # Log because asyncio would silently eat the error
        logger.exception("Broadcast error")
    return client_count


## Watcher thread


class PathIndex:
    """O(1) path lookup index for the flat FileEntry tree."""

    def __init__(self, root: list[FileEntry]):
        self.root = root
        self._index: dict[PurePosixPath, tuple[int, int]] = {}
        self._rebuild()

    def _rebuild(self):
        """Build path -> (start_idx, count) mapping in single O(n) pass."""
        index: dict[PurePosixPath, tuple[int, int]] = {}
        path_stack: list[tuple[PurePosixPath, int]] = []  # (path, start_idx)

        for i, entry in enumerate(self.root):
            # Pop completed paths from stack
            while path_stack and entry.level <= len(path_stack[-1][0].parts):
                completed_path, start_idx = path_stack.pop()
                index[completed_path] = (start_idx, i - start_idx)

            # Build current path
            if entry.level == 0:
                current_path = PurePosixPath()
            else:
                parent = path_stack[-1][0] if path_stack else PurePosixPath()
                current_path = parent / entry.name

            path_stack.append((current_path, i))

        # Close remaining paths
        for path, start_idx in path_stack:
            index[path] = (start_idx, len(self.root) - start_idx)

        self._index = index

    def get(self, path: PurePosixPath) -> tuple[int | None, list[FileEntry]]:
        """O(1) lookup: returns (start_idx, entries) or (None, [])."""
        if path not in self._index:
            return None, []
        start, count = self._index[path]
        return start, self.root[start : start + count]

    def find_insert_pos(self, path: PurePosixPath, isfile: int) -> int:
        """Find insertion position using index + binary search."""
        if not path.parts:
            return 0

        parent = path.parent
        name = path.name

        # Find parent's range
        if parent == PurePosixPath():
            # Insert at root level - scan root's direct children
            start, count = 0, len(self.root)
            target_level = 1
        elif parent in self._index:
            start, count = self._index[parent]
            start += 1  # Skip parent entry itself
            count -= 1
            target_level = len(parent.parts) + 1
        else:
            # Parent doesn't exist, shouldn't happen
            return len(self.root)

        # Binary search among direct children at target_level
        # Collect children indices first
        children = []
        i = start
        end = start + count
        while i < end:
            entry = self.root[i]
            if entry.level == target_level:
                children.append(i)
            i += 1

        if not children:
            return start

        # Binary search for insertion point
        nsort = sortkey(name)
        lo, hi = 0, len(children)
        while lo < hi:
            mid = (lo + hi) // 2
            idx = children[mid]
            entry = self.root[idx]
            ename = entry.name
            esort = sortkey(ename)
            # Compare: isfile, then sort key, then case-sensitive
            cmp = (
                entry.isfile - isfile
                or (esort > nsort) - (esort < nsort)
                or (ename > name) - (ename < name)
            )
            if cmp < 0:
                lo = mid + 1
            else:
                hi = mid

        if lo < len(children):
            return children[lo]
        elif children:
            # Insert after last child's subtree
            last_idx = children[-1]
            last_entry = self.root[last_idx]
            if last_entry.isfile:
                return last_idx + 1
            # Find end of last child's subtree
            last_path = parent / last_entry.name
            if last_path in self._index:
                s, c = self._index[last_path]
                return s + c
            return last_idx + 1
        return start

    def apply_update(
        self, path: PurePosixPath, new_entries: list[FileEntry]
    ) -> list[FileEntry]:
        """Apply an update and return the new root. Rebuilds index."""
        start, old_entries = self.get(path)

        if old_entries == new_entries:
            return self.root

        new_root = self.root[:]

        if start is not None:
            del new_root[start : start + len(old_entries)]

        if new_entries:
            # Rebuild index on modified list to find insert pos
            self.root = new_root
            self._rebuild()
            insert_pos = self.find_insert_pos(path, new_entries[0].isfile)
            new_root[insert_pos:insert_pos] = new_entries

        self.root = new_root
        self._rebuild()

        # Recalculate sizes for ancestor folders (including root)
        self._recalculate_ancestors(path)

        return new_root

    def _recalculate_ancestors(self, path: PurePosixPath):
        """Recalculate size/allocated for all ancestors of path, including root."""
        # Build list of ancestors from deepest to root
        ancestors = []
        current = path.parent if path.parts else PurePosixPath()
        while True:
            ancestors.append(current)
            if not current.parts:
                break
            current = current.parent

        # Process from deepest ancestor to root
        for ancestor_path in ancestors:
            if ancestor_path not in self._index:
                continue
            start, count = self._index[ancestor_path]
            if count == 0:
                continue

            ancestor = self.root[start]
            if ancestor.isfile:
                continue  # Files don't aggregate

            # Sum size/allocated of direct children
            total_size = 0
            total_allocated = 0
            i = start + 1
            while i < start + count:
                child = self.root[i]
                if child.level == ancestor.level + 1:
                    total_size += child.size
                    total_allocated += child.allocated
                    # Skip child's subtree
                    child_path = ancestor_path / child.name
                    if child_path in self._index:
                        _, child_count = self._index[child_path]
                        i += child_count
                    else:
                        i += 1
                else:
                    i += 1

            # Update ancestor entry if changed
            if ancestor.size != total_size or ancestor.allocated != total_allocated:
                self.root[start] = FileEntry(
                    level=ancestor.level,
                    name=ancestor.name,
                    key=ancestor.key,
                    mtime=ancestor.mtime,
                    size=total_size,
                    allocated=total_allocated,
                    isfile=ancestor.isfile,
                )


def collapse_paths(paths: set[PurePosixPath]) -> set[PurePosixPath]:
    """Remove child paths if parent is in set."""
    if not paths:
        return paths
    # Filter out root paths (empty or '.') which would cause full tree walks
    paths = {p for p in paths if p.parts and p.parts != (".",)}
    if not paths:
        return set()
    # Sort by depth (fewest parts first)
    sorted_paths = sorted(paths, key=lambda p: len(p.parts))
    result = set()
    for path in sorted_paths:
        # Check if any ancestor is already in result
        is_child = False
        for i in range(len(path.parts)):
            ancestor = PurePosixPath(*path.parts[:i]) if i > 0 else PurePosixPath()
            if ancestor in result:
                is_child = True
                break
        if not is_child:
            result.add(path)
    return result


# Debounce settings
DEBOUNCE_DELAY = 0.01  # Wait 10ms after last event
DEBOUNCE_MAX = 0.1  # But no more than 100ms total


def watcher(loop):
    """Unified watcher thread handling inotify, websocket signals, and periodic scans."""
    use_inotify = sys.platform == "linux"
    inotify_tree = None
    modified_flags = frozenset()

    if use_inotify:
        import inotify.adapters

        modified_flags = frozenset(
            (
                "IN_CREATE",
                "IN_DELETE",
                "IN_DELETE_SELF",
                "IN_MODIFY",
                "IN_MOVE_SELF",
                "IN_MOVED_FROM",
                "IN_MOVED_TO",
            )
        )

    while not quit.is_set():
        if use_inotify:
            import inotify.adapters

            inotify_tree = inotify.adapters.InotifyTree(rootpath.as_posix())

        # Initialize the tree from filesystem
        update_root(loop)
        path_index = PathIndex(state.root[:])

        trefresh = time.monotonic() + 300.0
        tspace = time.monotonic() + 5.0

        # Pending changes: path -> {"ws": count, "inotify": count}
        dirty_paths: dict[PurePosixPath, dict[str, int]] = {}
        first_event_time: float | None = None
        last_event_time: float | None = None

        def add_dirty(path: PurePosixPath, source: str) -> bool:
            """Add path to dirty set. Returns True if added, False if redundant."""
            nonlocal first_event_time, last_event_time
            # Check if already covered by an existing dirty path
            for existing in dirty_paths:
                if path == existing or (
                    len(path.parts) > len(existing.parts)
                    and path.parts[: len(existing.parts)] == existing.parts
                ):
                    # Count the event even if skipped
                    dirty_paths[existing][source] = (
                        dirty_paths[existing].get(source, 0) + 1
                    )
                    return False
            # Remove any paths that would be covered by this new one
            covered = {
                p
                for p in dirty_paths
                if len(p.parts) > len(path.parts)
                and p.parts[: len(path.parts)] == path.parts
            }
            # Aggregate counts from covered paths
            counts: dict[str, int] = {source: 1}
            for p in covered:
                for s, c in dirty_paths[p].items():
                    counts[s] = counts.get(s, 0) + c
                del dirty_paths[p]
            dirty_paths[path] = counts
            now = time.monotonic()
            if first_event_time is None:
                first_event_time = now
            last_event_time = now
            return True

        while not quit.is_set():
            now = time.monotonic()

            # Full refresh every 300s
            if now >= trefresh:
                break

            # Disk usage update every 5s
            if now >= tspace:
                tspace = now + 5.0
                update_space(loop)

            # Check if we should flush pending changes
            should_flush = False
            if dirty_paths:
                time_since_last = now - last_event_time if last_event_time else 0
                time_since_first = now - first_event_time if first_event_time else 0
                if (
                    time_since_last >= DEBOUNCE_DELAY
                    or time_since_first >= DEBOUNCE_MAX
                ):
                    should_flush = True

            if should_flush:
                paths_to_process = dirty_paths.copy()
                dirty_paths.clear()
                first_event_time = None
                last_event_time = None

                # Collapse paths (remove children if parent present)
                collapsed = collapse_paths(set(paths_to_process.keys()))

                # Process each collapsed path
                new_root = path_index.root
                for path in collapsed:
                    new_entries = walk(path)
                    new_root = path_index.apply_update(path, new_entries)

                # Broadcast if changed
                if new_root != state.root:
                    try:
                        update_msg = format_update(state.root, new_root)
                        with state.lock:
                            broadcast(update_msg, loop)
                            state.root = new_root
                    except Exception:
                        logger.exception("format_update failed; full rescan")
                        try:
                            fresh = walk(PurePosixPath())
                            path_index = PathIndex(fresh)
                            update_msg = format_update(state.root, fresh)
                            with state.lock:
                                broadcast(update_msg, loop)
                                state.root = fresh
                        except Exception:
                            logger.exception("Fallback failed; sending full root")
                            with state.lock:
                                broadcast(format_root(fresh), loop)
                                state.root = fresh

            # Collect events from websocket signals (non-blocking)
            try:
                while True:
                    path = _update_queue.get_nowait()
                    add_dirty(path, "ws")
            except queue.Empty:
                pass

            # Collect inotify events if available (short timeout for responsiveness)
            if inotify_tree:
                for event in inotify_tree.event_gen(yield_nones=False, timeout_s=0.05):
                    if quit.is_set():
                        return
                    if not (modified_flags & set(event[1])):
                        continue

                    # Extract relative path
                    path = PurePosixPath(event[2]) / event[3]
                    try:
                        rel_path = path.relative_to(rootpath)
                    except ValueError:
                        continue

                    # Skip dotfiles
                    if any(part.startswith(".") for part in rel_path.parts):
                        continue

                    add_dirty(rel_path, "inotify")

                    # Don't block too long collecting events
                    now = time.monotonic()
                    if first_event_time and now - first_event_time >= DEBOUNCE_MAX:
                        break
            else:
                # No inotify, just sleep briefly for responsiveness
                time.sleep(0.05)

        if inotify_tree:
            del inotify_tree


def start(app):
    global rootpath
    config.load_config()
    rootpath = config.config.path
    app.ctx.watcher = threading.Thread(
        target=watcher,
        args=[app.loop],
        # Descriptive name for system monitoring
        name=f"cista-watcher {rootpath}",
    )
    app.ctx.watcher.start()


def stop(app):
    quit.set()
    app.ctx.watcher.join()



from __future__ import annotations

import shutil
from pathlib import PurePosixPath
from typing import Any

import msgspec
from sanic import BadRequest

from cista import config
from cista.util import filename

## Control commands


class ControlBase(msgspec.Struct, tag_field="op", tag=str.lower):
    def __call__(self):
        raise NotImplementedError

    def affected_paths(self) -> list[str]:
        """Return list of paths affected by this operation for change notification."""
        return []


class MkDir(ControlBase):
    path: str

    def __call__(self):
        path = config.config.path / filename.sanitize(self.path)
        path.mkdir(parents=True, exist_ok=False)

    def affected_paths(self) -> list[str]:
        return [filename.sanitize(self.path)]


class Rename(ControlBase):
    path: str
    to: str

    def __call__(self):
        to = filename.sanitize(self.to)
        if "/" in to:
            raise BadRequest("Rename 'to' name should only contain filename, not path")
        path = config.config.path / filename.sanitize(self.path)
        path.rename(path.with_name(to))

    def affected_paths(self) -> list[str]:
        sanitized = filename.sanitize(self.path)
        new_path = str(PurePosixPath(sanitized).with_name(filename.sanitize(self.to)))
        return [sanitized, new_path]


class Rm(ControlBase):
    sel: list[str]

    def __call__(self):
        root = config.config.path
        sel = [root / filename.sanitize(p) for p in self.sel]
        for p in sel:
            if p.is_dir():
                shutil.rmtree(p)
            else:
                p.unlink()

    def affected_paths(self) -> list[str]:
        return [filename.sanitize(p) for p in self.sel]


class Mv(ControlBase):
    sel: list[str]
    dst: str

    def __call__(self):
        root = config.config.path
        sel = [root / filename.sanitize(p) for p in self.sel]
        dst = root / filename.sanitize(self.dst)
        if not dst.is_dir():
            raise BadRequest("The destination must be a directory")
        for p in sel:
            shutil.move(p, dst)

    def affected_paths(self) -> list[str]:
        dst = filename.sanitize(self.dst)
        paths = [filename.sanitize(p) for p in self.sel]
        # Include new locations in dst
        paths.extend(f"{dst}/{PurePosixPath(p).name}" for p in self.sel)
        return paths


class Cp(ControlBase):
    sel: list[str]
    dst: str

    def __call__(self):
        root = config.config.path
        sel = [root / filename.sanitize(p) for p in self.sel]
        dst = root / filename.sanitize(self.dst)
        if not dst.is_dir():
            raise BadRequest("The destination must be a directory")
        for p in sel:
            if p.is_dir():
                # Note: copies as dst rather than in dst unless name is appended.
                shutil.copytree(
                    p,
                    dst / p.name,
                    dirs_exist_ok=True,
                    ignore_dangling_symlinks=True,
                )
            else:
                shutil.copy2(p, dst)

    def affected_paths(self) -> list[str]:
        dst = filename.sanitize(self.dst)
        # Only destinations are new (sources unchanged)
        return [f"{dst}/{PurePosixPath(filename.sanitize(p)).name}" for p in self.sel]


ControlTypes = MkDir | Rename | Rm | Mv | Cp


## File uploads and downloads


class FileRange(msgspec.Struct):
    name: str
    size: int
    start: int
    end: int


class StatusMsg(msgspec.Struct):
    status: str
    req: FileRange


class ErrorMsg(msgspec.Struct):
    error: dict[str, Any]


## Directory listings


class FileEntry(msgspec.Struct, array_like=True, frozen=True):
    level: int
    name: str
    key: str
    mtime: int
    size: int
    allocated: int
    isfile: int

    def __str__(self):
        return self.key or "FileEntry()"

    def __repr__(self):
        return f"{self.name} ({self.size}, {self.mtime})"


class Update(msgspec.Struct, array_like=True): ...


class UpdKeep(Update, tag="k"):
    count: int


class UpdDel(Update, tag="d"):
    count: int


class UpdIns(Update, tag="i"):
    items: list[FileEntry]


class UpdateMessage(msgspec.Struct):
    update: list[UpdKeep | UpdDel | UpdIns]


class Space(msgspec.Struct):
    disk: int
    free: int
    used: int
    storage: int
    allocated: int



from __future__ import annotations

import os
import secrets
import sys
from contextlib import suppress
from functools import wraps
from hashlib import sha256
from pathlib import Path, PurePath
from time import sleep, time
from typing import Callable, Concatenate, Literal, ParamSpec

import msgspec
import msgspec.toml


class Config(msgspec.Struct):
    path: Path
    listen: str
    secret: str = secrets.token_hex(12)
    public: bool = False
    name: str = ""
    users: dict[str, User] = {}
    links: dict[str, Link] = {}


# Typing: arguments for config-modifying functions
P = ParamSpec("P")
ResultStr = Literal["modified", "created", "read"]
RawModifyFunc = Callable[Concatenate[Config, P], Config]
ModifyPublic = Callable[P, ResultStr]


class User(msgspec.Struct, omit_defaults=True):
    privileged: bool = False
    hash: str = ""
    lastSeen: int = 0  # noqa: N815


class Link(msgspec.Struct, omit_defaults=True):
    location: str
    creator: str = ""
    expires: int = 0


# Global variables - initialized during application startup
config: Config
conffile: Path


def init_confdir() -> None:
    global conffile
    if p := os.environ.get("CISTA_HOME"):
        home = Path(p)
    else:
        xdg = os.environ.get("XDG_CONFIG_HOME")
        home = (
            Path(xdg).expanduser() / "cista" if xdg else Path.home() / ".config/cista"
        )
    if not home.is_dir():
        home.mkdir(parents=True, exist_ok=True)
        home.chmod(0o700)
    conffile = home / "db.toml"


def derived_secret(*params, len=8) -> bytes:
    """Used to derive secret keys from the main secret"""
    # Each part is made the same length by hashing first
    combined = b"".join(
        sha256(p if isinstance(p, bytes) else f"{p}".encode()).digest()
        for p in [config.secret, *params]
    )
    # Output a bytes of the desired length
    return sha256(combined).digest()[:len]


def enc_hook(obj):
    if isinstance(obj, PurePath):
        return obj.as_posix()
    raise TypeError


def dec_hook(typ, obj):
    if typ is Path:
        return Path(obj)
    raise TypeError


def config_update(
    modify: RawModifyFunc,
) -> ResultStr | Literal["collision"]:
    global config
    tmpname = conffile.with_suffix(".tmp")
    try:
        f = tmpname.open("xb")
    except FileExistsError:
        if tmpname.stat().st_mtime < time() - 1:
            tmpname.unlink()
        return "collision"
    try:
        # Load, modify and save with atomic replace
        try:
            old = conffile.read_bytes()
            c = msgspec.toml.decode(old, type=Config, dec_hook=dec_hook)
        except FileNotFoundError:
            old = b""
            c = Config(path=Path(), listen="", secret=secrets.token_hex(12))
        c = modify(c)
        new = msgspec.toml.encode(c, enc_hook=enc_hook)
        if old == new:
            f.close()
            tmpname.unlink()
            config = c
            return "read"
        f.write(new)
        f.close()
        if sys.platform == "win32":
            # Windows doesn't support atomic replace
            with suppress(FileNotFoundError):
                conffile.unlink()
        tmpname.rename(conffile)  # Atomic replace
    except:
        f.close()
        tmpname.unlink()
        raise
    config = c
    return "modified" if old else "created"


def modifies_config(
    modify: Callable[Concatenate[Config, P], Config],
) -> Callable[P, ResultStr]:
    """Decorator for functions that modify the config file

    The decorated function takes as first arg Config and returns it modified.
    The wrapper handles atomic modification and returns a string indicating the result.
    """

    @wraps(modify)
    def wrapper(*args: P.args, **kwargs: P.kwargs) -> ResultStr:
        def m(c: Config) -> Config:
            return modify(c, *args, **kwargs)

        # Retry modification in case of write collision
        while (c := config_update(m)) == "collision":
            sleep(0.01)
        return c

    return wrapper


def load_config():
    global config
    init_confdir()
    raw = conffile.read_bytes()
    config = msgspec.toml.decode(raw, type=Config, dec_hook=dec_hook)
    # Migrate from old authentication field if present
    raw_dict = msgspec.toml.decode(raw)
    if "authentication" in raw_dict and "public" not in raw_dict:
        # Old config with authentication mode: migrate to public bool
        new_public = raw_dict["authentication"] == "none"
        config = msgspec.structs.replace(config, public=new_public)
        update_config({})  # Save the migrated config


@modifies_config
def update_config(conf: Config, changes: dict) -> Config:
    """Create/update the config with new values, respecting changes done by others."""
    # Encode into dict, update values with new, convert to Config
    settings = msgspec.to_builtins(conf, enc_hook=enc_hook)
    settings.update(changes)
    return msgspec.convert(settings, Config, dec_hook=dec_hook)


@modifies_config
def update_user(conf: Config, name: str, changes: dict) -> Config:
    """Create/update a user with new values, respecting changes done by others."""
    # Encode into dict, update values with new, convert to Config
    try:
        # Copy user by converting to dict and back
        u = msgspec.convert(
            msgspec.to_builtins(conf.users[name], enc_hook=enc_hook),
            User,
            dec_hook=dec_hook,
        )
    except KeyError:
        u = User()
    if "password" in changes:
        from . import auth

        auth.set_password(u, changes["password"])
        del changes["password"]
    udict = msgspec.to_builtins(u, enc_hook=enc_hook)
    udict.update(changes)
    settings = msgspec.to_builtins(conf, enc_hook=enc_hook)
    settings["users"][name] = msgspec.convert(udict, User, dec_hook=dec_hook)
    return msgspec.convert(settings, Config, dec_hook=dec_hook)


@modifies_config
def del_user(conf: Config, name: str) -> Config:
    """Delete named user account."""
    # Create a copy by converting to dict and back
    settings = msgspec.to_builtins(conf, enc_hook=enc_hook)
    settings["users"].pop(name)
    return msgspec.convert(settings, Config, dec_hook=dec_hook)



import asyncio
from contextlib import suppress


class AsyncLink:
    """Facilitate two-way connection between asyncio and a worker thread."""

    def __init__(self):
        """Initialize; must be called from async context."""
        self.loop = asyncio.get_running_loop()
        self.queue = asyncio.Queue(maxsize=1)

    async def __call__(self, command) -> asyncio.Future:
        """Run command in worker thread; awaitable.

        Args:
            command: Command to run in worker thread.
        """
        fut = self.loop.create_future()
        await self.queue.put((command, fut))
        return await fut

    @property
    def to_sync(self):
        """Yield SyncRequests from async caller when called from worker thread."""
        while (req := self._await(self._get())) is not None:
            yield SyncRequest(self, req)

    async def _get(self):
        """Retrieve an item from the queue; handle cancellation."""
        with suppress(asyncio.CancelledError):
            ret = await self.queue.get()
            self.queue.task_done()
            return ret

    def _await(self, coro):
        """Run coroutine in main thread and return result; called from worker."""
        return asyncio.run_coroutine_threadsafe(coro, self.loop).result()

    async def stop(self):
        """Stop worker and clean up."""
        while not self.queue.empty():
            command, future = self.queue.get_nowait()
            if not future.done():
                future.set_exception(Exception("AsyncLink stopped"))
            self.queue.task_done()
        await self.queue.put(None)


async def set_result(fut: asyncio.Future, value=None, exception=None):
    """Set result or exception on an asyncio.Future object.

    Args:
        fut (asyncio.Future): Future to set result or exception on.
        value: Result to set on the future.
        exception: Exception to set on the future.
    """
    with suppress(asyncio.InvalidStateError):
        if exception is None:
            fut.set_result(value)
        else:
            fut.set_exception(exception)


class SyncRequest:
    """Handle values from sync thread in main asyncio event loop."""

    def __init__(self, alink: AsyncLink, req):
        """Initialize SyncRequest with AsyncLink and request."""
        self.alink = alink
        self.command, self.future = req
        self.done = False

    def __enter__(self):
        """Provide command to with-block and handle exceptions."""
        return self.command

    def __exit__(self, exc_type, exc, traceback):
        """Set result or exception on exit; suppress exceptions in with-block."""
        if exc:
            self.set_exception(exc)
            return True
        if not self.done:
            self.set_result(None)
        return None

    def set_result(self, value):
        """Set result value; mark as done."""
        self.done = True
        self.alink._await(set_result(self.future, value))

    def set_exception(self, exc):
        """Set exception; mark as done."""
        self.done = True
        self.alink._await(set_result(self.future, exception=exc))



from time import time

import jwt

from cista.config import derived_secret


def session_secret():
    return derived_secret("session")


max_age = 365 * 86400  # Seconds since last login


def get(request):
    try:
        return jwt.decode(request.cookies.s, session_secret(), algorithms=["HS256"])
    except Exception:
        return False if "s" in request.cookies else None


def create(res, username, **kwargs):
    data = {
        "exp": int(time()) + max_age,
        "username": username,
        **kwargs,
    }
    s = jwt.encode(data, session_secret())
    res.cookies.add_cookie("s", s, httponly=True, max_age=max_age)


def update(res, s, **kwargs):
    s.update(kwargs)
    s = jwt.encode(s, session_secret())
    max_age = max(1, s["exp"] - int(time()))  # type: ignore
    res.cookies.add_cookie("s", s, httponly=True, max_age=max_age)


def delete(res):
    res.cookies.delete_cookie("s")


def flash(res, message: str | None):
    if message is None:
        res.cookies.delete_cookie("message")
    else:
        res.cookies.add_cookie("message", message, max_age=5)



from pathlib import Path

import msgspec


def readconf() -> dict:
    p = Path.home() / ".droppy/config"  # Hardcoded in Droppy
    cf = msgspec.json.decode((p / "config.json").read_bytes())
    db = msgspec.json.decode((p / "db.json").read_bytes())
    cf["path"] = p.parent / "files"
    cf["listen"] = _droppy_listeners(cf)
    return cf | db


def _droppy_listeners(cf):
    """Convert Droppy listeners to our format, for typical cases but not in full."""
    for listener in cf["listeners"]:
        try:
            if listener["protocol"] == "https":
                # TODO: Add support for TLS
                continue
            socket = listener.get("socket")
            if socket:
                if isinstance(socket, list):
                    socket = socket[0]
                return f"{socket}"
            port = listener["port"]
            if isinstance(port, list):
                port = port[0]
            host = listener["host"]
            if isinstance(host, list):
                host = host[0]
        except (KeyError, IndexError):
            continue
        else:
            if host in ("127.0.0.1", "::", "localhost"):
                return f":{port}"
            return f"{host}:{port}"

    # If none matched, fallback to Droppy default
    return "0.0.0.0:8989"



import hmac
import re
from time import time
from unicodedata import normalize

import argon2
import msgspec
from html5tagger import Document
from sanic import Blueprint, html, json, redirect
from sanic.exceptions import BadRequest, Forbidden, Unauthorized

from cista import config, session
from cista.util import pwgen

_LOGIN_PAGE_CSS = """\
/* ===========================================
   LOGIN PAGE STYLES
   Must match ModalDialog.vue global styles.
   =========================================== */
* { box-sizing: border-box; }
body {
    font-family: 'Roboto', system-ui, -apple-system, sans-serif;
    font-size: 1rem;
    margin: 0;
    min-height: 100vh;
    display: flex;
    align-items: center;
    justify-content: center;
    background: transparent;
}
.login-card {
    background: #ddd;
    color: #000;
    border-radius: 0.5rem;
    box-shadow: 0 0 1rem #0008;
    width: 100%;
    max-width: 320px;
}
h1 {
    background: #146;
    color: #fff;
    margin: 0;
    padding: 0.5rem 1rem;
    font-size: 1.2rem;
    font-weight: normal;
    border-radius: 0.5rem 0.5rem 0 0;
}
.content {
    padding: 1rem;
}
.message {
    color: #444;
    margin: 0 0 0.5rem 0;
    font-size: 0.875rem;
}
form {
    display: grid;
    grid-template-columns: auto 1fr;
    gap: 0.5rem 1rem;
    align-items: center;
}
label {
    font-size: 1rem;
}
input[type="text"],
input[type="password"] {
    font: inherit;
    font-size: 1rem;
    padding: 0.5rem;
    border: 2px solid #888;
    border-radius: 0.25rem;
    background: #fff;
    color: #000;
    min-width: 0;
}
input:focus {
    outline: none;
    border-color: #f80;
}
.button-row {
    grid-column: 1 / -1;
    display: flex;
    justify-content: flex-end;
    margin-top: 0.5rem;
}
button {
    font: inherit;
    font-size: 1rem;
    padding: 0.5rem 1rem;
    background: #146;
    color: #fff;
    border: none;
    border-radius: 0.25rem;
    cursor: pointer;
}
button:hover { background: #f80; }
button:disabled {
    background: #888;
    cursor: not-allowed;
}
.error {
    grid-column: 1 / -1;
    color: #c00;
    font-size: 0.875rem;
    min-height: 1.2em;
    margin: 0;
}
"""

_LOGIN_PAGE_JS = """\
const form = document.getElementById('loginForm');
const error = document.getElementById('error');
const submitBtn = document.getElementById('submitBtn');
const usernameField = document.getElementById('username');
const passwordField = document.getElementById('password');
const isInIframe = window.parent !== window;

// Focus username field on load
usernameField.focus();

const showError = (msg) => {
    error.textContent = msg;
    submitBtn.disabled = false;
    submitBtn.textContent = 'Log in';
    // Focus and select the relevant field
    if (msg.toLowerCase().includes('password')) {
        passwordField.focus();
        passwordField.select();
    } else {
        usernameField.focus();
        usernameField.select();
    }
};

form.onsubmit = async (e) => {
    e.preventDefault();
    error.textContent = '';
    submitBtn.disabled = true;
    submitBtn.textContent = 'Logging in...';

    try {
        const res = await fetch('/auth/login', {
            method: 'POST',
            headers: {
                'Content-Type': 'application/json',
                'Accept': 'application/json'
            },
            body: JSON.stringify({
                username: usernameField.value,
                password: passwordField.value
            })
        });

        if (res.ok) {
            if (isInIframe) {
                window.parent.postMessage({type: 'auth-success'}, '*');
            } else {
                window.location.href = '/';
            }
        } else {
            const data = await res.json();
            showError(data.message || data.detail || 'Login failed');
        }
    } catch (err) {
        showError('Connection error. Please try again.');
    }
};
"""

# Import for SSO validation (lazily loaded to avoid circular imports)
_sso_module = None


def _get_sso():
    global _sso_module
    if _sso_module is None:
        from cista import sso

        _sso_module = sso
    return _sso_module


_argon = argon2.PasswordHasher()
_droppyhash = re.compile(r"^([a-f0-9]{64})\$([a-f0-9]{8})$")


def _pwnorm(password):
    return normalize("NFC", password).strip().encode()


def login(username: str, password: str):
    un = _pwnorm(username)
    pw = _pwnorm(password)
    try:
        u = config.config.users[un.decode()]
    except KeyError:
        raise ValueError("Invalid username") from None
    # Verify password
    need_rehash = False
    if not u.hash:
        raise ValueError("Account disabled")
    if (m := _droppyhash.match(u.hash)) is not None:
        h, s = m.groups()
        h2 = hmac.digest(pw + s.encode() + un, b"", "sha256").hex()
        if not hmac.compare_digest(h, h2):
            raise ValueError("Invalid password")
        # Droppy hashes are weak, do a hash update
        need_rehash = True
    else:
        try:
            _argon.verify(u.hash, pw)
        except Exception:
            raise ValueError("Invalid password") from None
        if _argon.check_needs_rehash(u.hash):
            need_rehash = True
    # Login successful
    if need_rehash:
        set_password(u, password)
    now = int(time())
    u.lastSeen = now
    return u


def set_password(user: config.User, password: str):
    user.hash = _argon.hash(_pwnorm(password))


class LoginResponse(msgspec.Struct):
    user: str = ""
    privileged: bool = False
    error: str = ""


async def verify(request, *, privileged=False):
    """Verify that the request is authorized.

    For paskia mode (PASKIA_BACKEND_URL set), validates against the SSO backend.
    For built-in mode, checks session-based authentication.
    For public mode (config.public=True), skips auth unless privileged is required.

    Args:
        request: The Sanic request object
        privileged: If True, requires admin privileges (always enforced even in public mode)

    Raises:
        Unauthorized: If authentication is required
        Forbidden: If access is denied
    """
    # Public mode: skip auth unless privileged access is required
    if config.config.public and not privileged:
        return

    sso = _get_sso()
    if sso.paskia_enabled():
        perm = "cista:admin" if privileged else "cista:login"
        await sso.validate_sso_request(request, perm=perm)
        return

    user = getattr(request.ctx, "user", None)
    if privileged:
        if user and user.privileged:
            return
        raise Forbidden(
            "Access Forbidden: Only for privileged users",
            quiet=True,
        )
    if user:
        return
    raise Unauthorized(
        f"Login required for {request.path}",
        "cookie",
        context={"auth": {"iframe": "/auth/restricted/"}},
        quiet=True,
    )


# Blueprint for built-in auth (only registered when paskia is NOT enabled)
bp = Blueprint("auth", url_prefix="/auth")


@bp.get("/restricted/")
async def login_page(request):
    """Login page that works both standalone and in paskia iframe."""
    s = session.get(request)

    # Check if already logged in
    if s:
        # Already authenticated - signal success if in iframe
        return html(_login_success_page(s["username"]))

    doc = Document("Cista - Login")
    # Add paskia-compatible styling and scripts
    doc.style(_LOGIN_PAGE_CSS)
    with doc.div(class_="login-card"):
        doc.h1("Authentication Required")
        with doc.div(class_="content"):
            with doc.form(method="POST", id="loginForm", autocomplete="on"):
                doc.label("Username:", for_="username")
                doc.input(
                    type="text",
                    id="username",
                    name="username",
                    autocomplete="username webauthn",
                    required=True,
                )
                doc.label("Password:", for_="password")
                doc.input(
                    type="password",
                    id="password",
                    name="password",
                    autocomplete="current-password webauthn",
                    required=True,
                )
                with doc.div(class_="button-row"):
                    doc.button("Log in", type="submit", id="submitBtn")
                doc.p("", class_="error", id="error")

    # JavaScript for AJAX login and postMessage communication
    doc.script_(_LOGIN_PAGE_JS)

    res = html(doc)
    if s is False:
        session.delete(res)
    return res


def _login_success_page(username: str) -> str:
    """Minimal page that signals auth-success to parent iframe."""
    return str(
        Document().script_("window.parent.postMessage({type:'auth-success'},'*')")
    )


@bp.post("/login")
async def login_post(request):
    try:
        if request.headers.content_type == "application/json":
            username = request.json["username"]
            password = request.json["password"]
        else:
            username = request.form["username"][0]
            password = request.form["password"][0]
        if not username or not password:
            raise KeyError
    except KeyError:
        raise BadRequest(
            "Missing username or password",
            context={"redirect": "/login"},
        ) from None
    try:
        user = login(username, password)
    except ValueError as e:
        raise Forbidden(str(e), context={"redirect": "/login"}) from e

    if "text/html" in request.headers.accept:
        res = redirect("/")
        session.flash(res, "Logged in")
    else:
        res = json({"data": {"username": username, "privileged": user.privileged}})
    session.create(res, username)
    return res


@bp.post("/api/logout")
async def logout_post(request):
    s = request.ctx.session
    msg = "Logged out" if s else "Not logged in"
    if "text/html" in request.headers.accept:
        res = redirect("/login")
        res.cookies.add_cookie("flash", msg, max_age=5)
    else:
        res = json({"message": msg})
    session.delete(res)
    return res


@bp.post("/password-change")
async def change_password(request):
    try:
        if request.headers.content_type == "application/json":
            username = request.json["username"]
            pwchange = request.json["passwordChange"]
            password = request.json["password"]
        else:
            username = request.form["username"][0]
            pwchange = request.form["passwordChange"][0]
            password = request.form["password"][0]
        if not username or not password:
            raise KeyError
    except KeyError:
        raise BadRequest(
            "Missing username, passwordChange or password",
        ) from None
    try:
        user = login(username, password)
        set_password(user, pwchange)
    except ValueError as e:
        raise Forbidden(str(e), context={"redirect": "/login"}) from e

    if "text/html" in request.headers.accept:
        res = redirect("/")
        session.flash(res, "Password updated")
    else:
        res = json({"message": "Password updated"})
    session.create(res, username)
    return res


@bp.get("/users")
async def list_users(request):
    await verify(request, privileged=True)
    users = []
    for name, user in config.config.users.items():
        users.append(
            {
                "username": name,
                "privileged": user.privileged,
                "lastSeen": user.lastSeen,
            }
        )
    return json({"users": users})


@bp.post("/users")
async def create_user(request):
    await verify(request, privileged=True)
    try:
        if request.headers.content_type == "application/json":
            username = request.json["username"]
            password = request.json.get("password")
            privileged = request.json.get("privileged", False)
        else:
            username = request.form["username"][0]
            password = request.form.get("password", [None])[0]
            privileged = request.form.get("privileged", ["false"])[0].lower() == "true"
        if not username or not username.isidentifier():
            raise ValueError("Invalid username")
    except (KeyError, ValueError) as e:
        raise BadRequest(str(e)) from e
    if username in config.config.users:
        raise BadRequest("User already exists")
    if not password:
        password = pwgen.generate()
    changes = {"privileged": privileged}
    changes["hash"] = _argon.hash(_pwnorm(password))
    try:
        config.update_user(username, changes)
    except Exception as e:
        raise BadRequest(str(e)) from e
    return json({"message": f"User {username} created", "password": password})


@bp.put("/users/<username>")
async def update_user(request, username):
    await verify(request, privileged=True)
    try:
        if request.headers.content_type == "application/json":
            changes = request.json
        else:
            changes = {}
            if "password" in request.form:
                changes["password"] = request.form["password"][0]
            if "privileged" in request.form:
                changes["privileged"] = request.form["privileged"][0].lower() == "true"
    except KeyError as e:
        raise BadRequest("Missing fields") from e
    password_response = None
    if "password" in changes:
        if changes["password"] == "":
            changes["password"] = pwgen.generate()
        password_response = changes["password"]
        changes["hash"] = _argon.hash(_pwnorm(changes["password"]))
        del changes["password"]
    if not changes:
        return json({"message": "No changes"})
    try:
        config.update_user(username, changes)
    except Exception as e:
        raise BadRequest(str(e)) from e
    response = {"message": f"User {username} updated"}
    if password_response:
        response["password"] = password_response
    return json(response)


@bp.delete("/users/<username>")
async def delete_user(request, username):
    await verify(request, privileged=True)
    if username not in config.config.users:
        raise BadRequest("User does not exist")
    try:
        config.del_user(username)
    except Exception as e:
        raise BadRequest(str(e)) from e
    return json({"message": f"User {username} deleted"})



from sanic import Sanic, exceptions, response

app = Sanic("server80")


# Send all HTTP users to HTTPS
@app.exception(exceptions.NotFound, exceptions.MethodNotSupported)
def redirect_everything_else(request, exception):
    server, path = request.server_name, request.path
    if server and path.startswith("/"):
        return response.redirect(f"https://{server}{path}", status=308)
    return response.text("Bad Request. Please use HTTPS!", status=400)


# ACME challenge for LetsEncrypt
@app.get("/.well-known/acme-challenge/<challenge>")
async def letsencrypt(request, challenge):
    try:
        return response.text(acme_challenges[challenge])
    except KeyError:
        return response.text(f"ACME challenge not found: {challenge}", status=404)


acme_challenges = {}
